
/*
 *
 *
 *                 #                 #
 *                 #                 #
 *       ####    #####     ####    #####
 *      #    #     #           #     #
 *       ##        #       #####     #
 *         ##      #      #     #    #
 *      #    #     #  #   #    ##    #  #
 *       ####       ##     #### #     ##
 *
 *
 *	"display formatted file info"
 *
 *	             or
 *
 *	     "a doit yourself ls"
 *
 *
 *
 *	usage:   stat [-VnrwxtugoiUGsamc] [-f arg] file ...
 *
 *	Author:
 *		Goran Larsson              hacker@isadora.ikp.liu.se
 *		Jonn O Nilsons Pl 4        ...!enea!liuida!prodix!isadora!hacker
 *		S-582 70  LINKOPING
 *		SWEDEN
 *
 *	This program is released in the public domain, give it away
 *	but do not take any money for it!
 *
 */

#define VERSION "2.4"
 
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <stdio.h>
#include <pwd.h>
#include <grp.h>
#include <time.h>
#include "config.h"

#define MAXPATH 1024		/* must be >= MAXPATHLEN in BSD */
#define MAXFORMAT 1024
#define MAXWIDTH 128
#define FUZZ 16

struct passwd *getpwuid ();
struct group  *getgrgid ();

char *ME;
char outbuffer[BUFSIZ];
char m[MAXWIDTH+FUZZ];

/* from libc.a */

int strlen ();
int strcmp ();
char *strcpy ();
char *strcat ();

#ifdef STRRCHR
  char *strrchr ();
#endif

#ifdef RINDEX
  char *rindex ();
# define strrchr(a,b) rindex(a,b)
#endif

/* from getopt.c or libc.a */

int getopt ();
int optind;
char *optarg;

/* this file */

int makefull ();
int display ();
char *putprot ();
int putmode ();
int Fnorm ();

char me[]      = "stat";
char str[]     = "%s";
char tstr[]    = "%.24s";
char dec[]     = "%d";
char sixstar[] = "****** ";

usage ()
{
	fprintf (
	  stderr,
	  "usage: %s [-VnNrwxtugoiUGsamc] [-f arg] file..\n",
	  ME
	);
	fflush ( stdout );
	exit ( 1 );
}

main ( argc, argv )
int argc;
char *argv[];

{
	struct stat buf;
	char fyle[MAXPATH+1];
	int c, i, j, lastopt;
	char expanded = 0;
	char format[MAXFORMAT];
	char formatted = 0;
	int field;
	char infield = 0;
	char align;
	char *pbuf;

	if ( ME = strrchr ( argv[0], '/' ) ) {
		ME++;
	} else {
		ME = argv[0];
	}

	setbuf ( stdout, outbuffer );

	while ( ( c = getopt ( argc, argv, "VnNrwxtugoiUGsamcf:" ) ) != EOF ) {
		switch ( lastopt = c ) {
			case 'V' :
				printf ( ME );
				if ( strcmp ( ME, me ) != 0 ) {
					printf ( " (%s)", me );
				}
				printf ( " version %s\n", VERSION );
				fflush ( stdout );
				break;
			case 'n' :
			case 'N' :
			case 'r' :
			case 'w' :
			case 'x' :
			case 't' :
			case 'u' :
			case 'g' :
			case 'o' :
			case 'i' :
			case 'U' :
			case 'G' :
			case 's' :
			case 'a' :
			case 'm' :
			case 'c' :
				format[formatted++] = '%';
				format[formatted++] = c;
				format[formatted++] = ' ';
				format[formatted+1] = '\0';
				break;
			case 'f' :
				for ( i = 0; i < strlen ( optarg ); i++ ) {
					format[formatted++] = optarg[i];
					format[formatted+1] = '\0';
				}
				break;
			case '?' :
				usage ();
		}
	}

	if ( formatted && ( lastopt != 'f' ) ) {
		format[formatted++] = '\\';
		format[formatted++] = 'n';
		format[formatted+1] = '\0';
	}

	if ( optind >= argc ) {
		usage ();
	}

	for ( i = optind; i < argc; i++ ) {

		expanded = 0;

		if ( !formatted && ( i > optind ) ) {
			printf ( "----------------------------\n" );
		}

		for ( j = 0; j < strlen ( argv[i] ); j++ ) {
			fyle[j] = argv[i][j];
			fyle[j+1] = '\0';
		}

		/*
		** probe file
		*/

		if ( stat ( fyle, &buf )) {
			fprintf ( stderr, "%s: cannot stat %s\n", ME, fyle );
			continue; /* with next file */
		}

		/*
		** handle formatted output
		*/

		if ( formatted ) {
		    for ( j = 0; j < formatted; j++ ) {
		    	if ( infield ) {
		    	    switch ( c = format[j] ) {
				case '0' :
				case '1' :
				case '2' :
				case '3' :
				case '4' :
				case '5' :
				case '6' :
				case '7' :
				case '8' :
				case '9' :
				    field = ( field * 10 ) + ( c - '0' );
				    break;
				case 'l' :
				case 'r' :
				case 'c' :
				    align = c;
				    goto postfield;	/* sorry */
				    break;
				default :
				    align = c;
				    goto postfield;	/* sorry */
			    }
		    	    continue;
		    	}
			switch ( c = format[j] ) {
			    case '%' :
				field = 0;
postfield:			infield = 0;
				switch ( c = format[++j] ) {
				    case 'n' :
					sprintf ( m, str, argv[i] );
					display ( m, field, align );
					break;
				    case 'N' :
					if ( !expanded ) {
					    makefull ( fyle );
					    expanded++;
					}
					sprintf ( m, str, fyle );
					display ( m, field, align );
					break;
				    case 'r' :
				    case 'w' :
				    case 'x' :
				    case 't' :
				    case 'u' :
				    case 'g' :
				    case 'o' :
					pbuf = putprot ( c, buf.st_mode );
					display ( pbuf, field, align );
					break;
				    case 'i' :
					sprintf ( m, dec, buf.st_ino );
					display ( m, field, align );
					break;
				    case 'U' :
					sprintf ( m, dec, buf.st_uid );
					display ( m, field, align );
					break;
				    case 'G' :
					sprintf ( m, dec, buf.st_gid );
					display ( m, field, align );
					break;
				    case 's' :
					sprintf ( m, "%d", buf.st_size );
					display ( m, field, align );
					break;
				    case 'a' :
					sprintf (
					    m,
					    tstr,
					    ctime ( &buf.st_atime )
					);
					display ( m, field, align );
					break;
				    case 'm' :
					sprintf (
					    m,
					    tstr,
					    ctime ( &buf.st_mtime )
					);
					display ( m, field, align );
					break;
				    case 'c' :
					sprintf (
					    m,
					    tstr,
					    ctime ( &buf.st_ctime )
					);
					display ( m, field, align );
					break;
				    case '0' :
				    case '1' :
				    case '2' :
				    case '3' :
				    case '4' :
				    case '5' :
				    case '6' :
				    case '7' :
				    case '8' :
				    case '9' :
					field = c - '0';
					infield = 1;
					break;
				    case '%' :
					fputc ( '%', stdout );
					break;
				    default :
					if ( field ) {
					    printf (
					      "%%%d%c%c", field, align, c
					    );
					} else {
					    printf ( "%%%c", c );
					}
					break;
				}
				break;
			    case '\\' :
				switch ( c = format[++j] ) {
				    case 'n' :
					fputc ( '\n', stdout );
					break;
				    case 't' :
					fputc ( '\t', stdout );
					break;
				    case '%' :
					fputc ( '%', stdout );
					break;
				    case '\\' :
					fputc ( '\\', stdout );
					break;
				    default  :
					printf ( "\\%c", c );
					break;
				}
				break;
			    default  :
				fputc ( c, stdout );
				break;
			    }
			}
			fflush ( stdout );
			continue; /* with next file */
		}

		/*
		** handle full output (no switches)
		*/

		makefull ( fyle );
		printf ( "\nFile: %s\n", fyle );

		printf ( "File mode          :  %o \t", buf.st_mode );
		putmode ( buf.st_mode );
		printf ( "\n" );

		printf ( "Inode number       :  %d\n", buf.st_ino );

		printf ( "Device ID          :  %d \t%d,%d\n",
			buf.st_dev,
			( buf.st_dev >> 8 ) & 0xFF,
			buf.st_dev & 0xFF
		);

		printf ( "Special device ID  :  %d \t%d,%d\n",
			buf.st_rdev,
			( buf.st_rdev >> 8 ) & 0xFF,
			buf.st_rdev & 0xFF
		);

		printf ( "Number of links    :  %d\n", buf.st_nlink );

		if ( getpwuid ( buf.st_uid ) == 0 ) {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				sixstar
			);
		} else {
			printf ( "User ID            :  %d \t(%s)\n",
				buf.st_uid,
				getpwuid ( buf.st_uid ) -> pw_name
			);
		}

		if ( getgrgid ( buf.st_gid ) == 0 ) {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				sixstar
			);
		} else {
			printf ( "Group ID           :  %d \t(%s)\n",
				buf.st_gid,
				getgrgid ( buf.st_gid ) -> gr_name
			);
		}

		printf ( "Size in bytes      :  %d\n", buf.st_size );

		printf ( "Last access        :  %s",
			ctime ( &buf.st_atime )
		);

		printf ( "Last modification  :  %s",
			ctime ( &buf.st_mtime )
		);

		printf ( "Last status change :  %s",
			ctime ( &buf.st_ctime )
		);

		printf ( "\n" );
		fflush ( stdout );
	}

	exit ( 0 );
}

char path[MAXPATH+1];

makefull ( fyle )
char *fyle;

{
	char work[MAXPATH+1];

	if ( fyle[0] != '/' ) {
		whereami ();
		if ( ( strlen ( path ) + 1 + strlen ( fyle ) ) > MAXPATH ) {
			fprintf ( stderr, "%s: path too long!\n", ME );
			fflush ( stdout );
			exit ( 1 );
		}
		strcpy ( work, path );
		strcat ( work, "/"  );
		strcat ( work, fyle );
		strcpy ( fyle, work );
	}
	Fnorm ( fyle );
}

#ifdef GETCWD
whereami ()		/* we have getcwd(), probably sys V */
{
	static int done_getcwd = 0;
	extern int getcwd ();

	char work[MAXPATH+1];

	if ( !done_getcwd ) {
		if ( getcwd ( path, MAXPATH ) <= 0 ) {
			fprintf ( stderr, "%s: cwd too long\n", ME );
			fflush ( stdout );
			exit ( 1 );
		}
		done_getcwd++;
	}
}
#endif

#ifdef GETWD
whereami ()		/* we have getwd(), probably late BSD */
{
	static int done_getwd = 0;
	extern char *getwd ();

	char work[MAXPATH+1];

	if ( !done_getwd ) {
		if ( getwd ( path ) == ((char *)0) ) {
			fprintf ( stderr, "%s: %s\n", ME, path );
			fflush ( stdout );
			exit ( 1 );
		}
		done_getwd++;
	}
}
#endif

#ifndef GETCWD
# ifndef GETWD
whereami ()		/* no getcwd(), no getwd(), old system */
{
	static int done_pwd = 0;
	extern FILE *popen ();
	extern int pclose ();

	FILE *fp;
	int i;
	char *slave = "pwd";	/* usually /bin/pwd or shell builtin */

	if ( !done_pwd ) {			/* THE HARD WAY */
		fp = popen ( slave, "r" );
		if ( !fp ) {
			printf ( "%s: can not execute %s\n", ME, slave );
			exit ( 1 );
		}
		for ( i = 0; ; i++ ) {
			if ( ( path[i] = getc ( fp ) ) == -1 ) {
				path[i-1] = '\0';
				break;
			}
		}
		pclose ( fp );
		done_pwd++;
	}
}
# endif
#endif

Fnorm ( cp )
char *cp;

{
	register char *p, *sp;
	register int slash;

	/* assume name begins with a '/' */

	for ( p = cp; *p; ) {		/* for each component */
		sp = p;			/* save slash address */
		while(*++p == '/') {	/* flush extra slashes */
			/* void */ ;
		}
		if ( p != ++sp ) {
			strcpy ( sp, p );
		}
		p = sp;			/* save start of component */
		slash = 0;
		while ( *++p ) {	/* find next slash or end of path */
			if ( *p == '/' ) {
				slash = 1;
				*p = 0;
				break;
			}
		}
		if ( *sp == '\0' ) {	/* if component is null */
			if ( --sp == cp ) { /* if path is one char (i.e. /) */
				break;
			} else {
				*sp = '\0';
			}
		} else {
			if ( strcmp ( ".", sp ) == 0 ) {
				if ( slash ) {
					strcpy ( sp, ++p );
					p = --sp;
				} else {
					if ( --sp != cp ) {
						*sp = '\0';
					}
				}
			} else {
				if ( strcmp ( "..", sp ) == 0 ) {
					if ( --sp != cp ) {
						while ( *--sp != '/' ) {
						}
					}
					if ( slash ) {
						strcpy ( ++sp, ++p );
						p = --sp;
					} else {
						if ( cp == sp ) {
							*++sp = '\0';
						} else {
							*sp = '\0';
						}
					}
				} else {
					if ( slash ) {
						*p = '/';
					}
				}
			}
		}
	}
}

display ( buf, width, how )
char *buf;
int width;
char how;

{
	register int i, j, sw;
	int trunc;

	trunc = 0;
	sw = strlen ( buf );
	if ( width > MAXWIDTH ) {		/* field to long ?         */
		printf ( "**FTL**" );		/* ERR = "Field To Long"   */
		return;
	}
	if ( width == 0 ) {			/* default justification ? */
		width = sw;
		how = 'l';
	}
	if ( sw > width ) {			/* item to long ?          */
		trunc = sw - width;
		sw = width;
	}
	switch ( how ) {
		case 'l' :				/* left justify    */
			for ( i = 0; i < sw; i++ ) {
				if ( ( i == ( sw-1 ) ) && trunc ) {
					buf[i] = '!';
				}
				fputc ( buf[i], stdout );
			}
			i = width - sw;
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			break;
		case 'r' :				/* right justify   */
			i = width - sw;
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			for ( i = 0; i < sw; i++ ) {
				if ( ( i == 0 ) && trunc ) {
					buf[trunc+i] = '!';
				}
				fputc ( buf[trunc+i], stdout );
			}
			break;
		case 'c' :				/* center          */
			i = ( width - sw ) / 2;
			j = width - i - sw;
			if ( trunc ) {
				trunc = trunc / 2;
				if ( trunc ) {
					buf[trunc] = '!';
				}
				buf[trunc+width-1] = '!';
			}
			while ( i-- > 0 ) {
				fputc ( ' ', stdout );
			}
			for ( i = 0; i < sw; i++ ) {
				fputc ( buf[trunc+i], stdout );
			}
			while ( j-- > 0 ) {
				fputc ( ' ', stdout );
			}
			break;
		default :
			printf ( "**BFJ**" );	/* ERR = "Bad Format Just." */
			break;
	}
}

char *putprot ( type, mode )
char type;
int mode;

{
	int i, j, shif, mask;
	static char tmp[4];

	j = mask = 0;
	switch ( type ) {
		case 't' :
			switch ( ( mode >> 12 ) & 0xF ) {
				case 010 : tmp[0] = '-'; break;	/* FREG */
				case 004 : tmp[0] = 'd'; break;	/* FDIR */
				case 002 : tmp[0] = 'c'; break;	/* FCHR */
				case 006 : tmp[0] = 'b'; break;	/* FBLK */
				case 007 : tmp[0] = '7'; break;	/* FMPB */
				case 003 : tmp[0] = '3'; break;	/* FMPC */
				case 005 : tmp[0] = '5'; break;	/* FNAM */
				case 001 : tmp[0] = 'p'; break;	/* FIFO */
				default:   tmp[0] = '?'; break;
			}
			tmp[1] = '\0';
			return ( &tmp[0] );
		case 'r' : mask = 4; break;
		case 'w' : mask = 2; break;
		case 'x' : mask = 1; break;
	}
	if ( mask ) {
		for ( i = 2; i >= 0; i-- ) {
			if ( ( ( mode >> ( i * 3 ) ) & 0x7 ) & mask ) {
				switch ( i ) {
					case 2: tmp[j++] = 'u'; break;
					case 1: tmp[j++] = 'g'; break;
					case 0: tmp[j++] = 'o'; break;
				}
			} else {
				tmp[j++] = ' ';
			}
		}
		tmp[j] = '\0';
		return ( &tmp[0] );
	}
	switch ( type ) {
		case 'u' : shif = 2; break;
		case 'g' : shif = 1; break;
		case 'o' : shif = 0; break;
	}
	for ( i = 2; i >= 0; i-- ) {
		if ( ( mode >> ( ( shif * 3 ) + i ) ) & 1 ) {
			switch ( i ) {
				case 2: tmp[j++] = 'r'; break;
				case 1: tmp[j++] = 'w'; break;
				case 0: tmp[j++] = 'x'; break;
			}
		} else {
			tmp[j++] = '-';
		}
	}
	tmp[j] = '\0';
	return ( &tmp[0] );
}

putmode ( mode )
int mode;
{
	int i;

	for ( i = 2; i >= 0; i-- ) {
		switch ( i ) {
			case 2: printf ( "u:" ); break;
			case 1: printf ( "g:" ); break;
			case 0: printf ( "o:" ); break;
		}
		switch ( ( mode >> ( i * 3 ) ) & 0x7 ) {
			case 000 : printf ( "--- " ); break;
			case 001 : printf ( "--x " ); break;
			case 002 : printf ( "-w- " ); break;
			case 003 : printf ( "-wx " ); break;
			case 004 : printf ( "r-- " ); break;
			case 005 : printf ( "r-x " ); break;
			case 006 : printf ( "rw- " ); break;
			case 007 : printf ( "rwx " ); break;
		}
	}

	switch ( ( mode >> 12 ) & 0xF ) {
		case 010 : printf ( "file   " ); break;		/* FREG */
		case 004 : printf ( "dir    " ); break;		/* FDIR */
		case 002 : printf ( "char   " ); break;		/* FCHR */
		case 006 : printf ( "block  " ); break;		/* FBLK */
		case 007 : printf ( "*FMPB* " ); break;		/* FMPB */
		case 003 : printf ( "*FMPC* " ); break;		/* FMPC */
		case 005 : printf ( "*FNAM* " ); break;		/* FNAM */
		case 001 : printf ( "pipe   " ); break;		/* FIFO */
		default:   printf ( sixstar   ); break;
	}

	switch ( ( mode >> 10 ) & 0x3 ) {
		case 000 : printf ( "        " ); break;
		case 001 : printf ( "GID     " ); break;	/* SGID */
		case 002 : printf ( "UID     " ); break;	/* SUID */
		case 003 : printf ( "GID UID " ); break;
	}

	switch ( ( mode >> 9 ) & 0x1 ) {
		case 000 : printf ( "       " ); break;
		case 001 : printf ( "*SVTX* " ); break;		/* SVTX */
	}
}
