
/*
 *
 *	init.h
 *
 *	Definitions for interface to HNIX init process
 *
 *	Author:
 *	     _
 *	    Goran Larsson
 *	    hacker@isadora.ikp.liu.se
 *	                           _
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                        _                                          *
 * Copyright (C) 1989 by Goran Larsson           All rights reserved *
 *                                                                   *
 * This is an unpublished work containing confidential information   *
 * that is the property of Goran Larsson.                            *
 * Any unauthorized use, duplication or disclosure is prohibited.    *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 */

static char RCSid_init_h[] = "$Header: init.h,v 1.2 89/03/12 19:11:42 root Exp $";

/*
 *	$Log:	init.h,v $
 * Revision 1.2  89/03/12  19:11:42  root
 * Added more comments about init states.
 * 
 * Revision 1.1  89/03/11  22:41:22  root
 * Initial revision
 * 
 */

#ifndef _INIT_
#define _INIT_

/*
** operations
*/

#define INIT_OPR_STATE  0	/* return current init state	*/
#define INIT_OPR_UNLOCK 1	/* remove unwanted lock		*/
#define INIT_OPR_CMD    2	/* issue command		*/

/*
** commands
*/

#define INIT_MERGE   0		/* read /etc/ttys		*/
#define INIT_SINGLE  1		/* enter singleuser mode	*/
#define INIT_HALT    2		/* halt system			*/
#define INIT_HALTN   3		/* halt without sync()		*/
#define INIT_INHIBIT 4		/* inhibit further logins	*/
#define INIT_ENABLE  5		/* enable further logins	*/
#define INIT_NMBR_CMDS 6

/*
** result codes
*/

#define INIT_OK	      0		/* all is well			*/
#define INIT_ERR_SHM  1		/* can not attach shared memory	*/
#define INIT_ERR_BUSY 2		/* init busy!			*/
#define INIT_ERR_CMD  3		/* bad command			*/
#define INIT_ERR_OPR  4		/* bad operation		*/
#define INIT_ERR_DONE 5		/* init did not signal done	*/

/*
** structure of shared memory
**
** Lock is true if another process has locked init.
** Command is one of the commands listed above.
** Busy is true if init is busy and can't accept a command.
** Done is true when init has completed your command.
** State is one of:
**	?	intermediate or unknown
**	S	single user mode
**	r	running /etc/rc first time
**	R	running /etc/rc following times (multi -> single -> multi)
**	M	multiuser mode
**	m	multiuser mode with no new getty's
**	D	shutting down
**	H	about to halt
**
** State changes case I		boot via single user mode and then shutdown
**
**	?S <EOF> rM <shutdown -h> mDH
**
** State changes case II	boot directly to multiuser mode and shutdown
**
**	?rM <shutdown -h> mDH
**
** State changes case III	boot via single user mode, visit single user
**				later, and then shutdown
**
**	?S <EOF> rM <shutdown> DS <EOF> RM <shutdown -h> mDH
**
*/

struct shared {
 	char lock;		/* init has been locked			*/
	char command;		/* command for init to process		*/
	char busy;		/* init is processing a command		*/
	char done;		/* init done with command when true	*/
	char state;		/* current state of init		*/
};

/*
** int init ( operation, command )
** int operation;
** int command;
*/

extern int init ();

#endif _INIT_
