
/*
 *
 *                    ##    #
 *   #        #  #     #  # #
 *               #     #    #
 *   # # ###  # ###    #  # # ###
 *   # ##   # #  #     #  # ##   #
 *   # #    # #  #     #  # #    #
 *   # #    # #  #     #  # #    #
 *   # #    # #  #  #  #  # ##   #
 *   # #    # #   ##  ### # # ###
 *
 *
 *	Interface to HNIX init process
 *
 *	Author:
 *	     _
 *	    Goran Larsson
 *	    hacker@isadora.ikp.liu.se
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                        _                                          *
 * Copyright (C) 1989 by Goran Larsson           All rights reserved *
 *                                                                   *
 * This is an unpublished work containing confidential information   *
 * that is the property of Goran Larsson.                            *
 * Any unauthorized use, duplication or disclosure is prohibited.    *
 *                                                                   *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * Note: Do not use this if you don't know what you want to do,
 *       and know how to do it. Playing around with init may do
 *       strange things to your system.
 *
 */

static char RCSid[] = "$Header: initlib.c,v 1.2 89/03/12 19:12:10 root Exp $";

/*
 *	$Log:	initlib.c,v $
 * Revision 1.2  89/03/12  19:12:10  root
 * Added more comments.
 * 
 * Revision 1.1  89/03/11  22:41:55  root
 * Initial revision
 * 
 */

#include <sys/types.h>
#include <sys/ipc.h>
#include <signal.h>

#include "init.h"

#define FALSE 0
#define TRUE  1

#define KEY 0x49	/* Init */

int
init ( operation, command )
int operation;
int command;

{
	static int shmid = -1;
	static struct shared *shm;

	int timeouts;
	int retries;
	int (*oldsig) ();

	/*
	** map shared memory if not mapped yet
	*/

	if ( shmid < 0 ) {
		shmid = shmget ( (key_t)KEY, sizeof (struct shared), 0777 );
		if ( shmid < 0 ) {
			return INIT_ERR_SHM;
		}
		shm = (struct shared *)shmat ( shmid, (char *)0, 0 );
	}

	/*
	** different operations are possible
	*/

	switch ( operation ) {

		/*
		** a bad operation
		*/

	default:
		return INIT_ERR_OPR;

		/*
		** a request for the state of init
		*/

	case INIT_OPR_STATE:
		return shm->state;

		/*
		** a request to remove an unwanted init lock
		*/

	case INIT_OPR_UNLOCK:
		shm->lock = FALSE;
		return INIT_OK;

		/*
		** a request to make init do something
		*/

	case INIT_OPR_CMD:
		if ( command < 0 || command > INIT_NMBR_CMDS ) {
			return INIT_ERR_CMD;
		}

		/*
		** wait until init is not busy, nor locked
		*/
	
		oldsig = signal ( SIGALRM, SIG_DFL );
		timeouts = 100;					/* 10 seconds */
		while ( timeouts-- && shm->busy || shm->lock ) {
			nap ( 100 );
		}
		if ( shm->busy ) {
			return INIT_ERR_BUSY;
		}
	
		/*
		** lock init and place command in shared memory
		*/
	
		shm->lock = TRUE;
		shm->command = command;
		shm->done = FALSE;
	
		/*
		** tell init to execute the command
		*/
	
		retries = 3;
		while ( retries-- ) {
			(void) kill ( 1, SIGINT );
			timeouts = 20;				/* 4 seconds */
			while ( timeouts-- && !shm->done ) {
				nap ( 200 );
			}
			if ( shm->done ) {
				break;
			}
		}
		if ( !shm->done ) {
			shm->done = FALSE;
			shm->lock = FALSE;
			return INIT_ERR_DONE;
		}
		shm->done = FALSE;
		shm->lock = FALSE;
		(void) signal ( SIGALRM, oldsig );
		return INIT_OK;
	}
}
